@@ -19,6 +19,8 @@
 #include <osmocom/core/fsm.h>
 #include <osmocom/core/tdef.h>
 #include <osmocom/core/time_cc.h>
+#include <osmocom/core/linuxlist.h>
+#include <osmocom/core/linuxrbtree.h>
 
 #include <osmocom/crypt/auth.h>
 
@@ -51,6 +53,9 @@
  * with delta = time between expiration of T3124 and receiving HANDOVER FAILURE by the serving BSC. */
 #define GSM_NY1_DEFAULT ((unsigned long)((GSM_T3124_MAX + GSM_NY1_REQ_DELTA)/GSM_T3105_DEFAULT + 1))
 
+#define SCCP_CONN_ID_UNSET	0xFFFFFFFF
+#define SCCP_CONN_ID_MAX	0x00FFFFFE
+
 struct mgcp_client_conf;
 struct mgcp_client;
 struct gsm0808_cell_id;
@@ -137,6 +142,7 @@ struct assignment_request {
 	enum assign_for assign_for;
 
 	bool aoip;
+	bool vgcs;
 
 	uint16_t msc_assigned_cic;
 
@@ -163,6 +169,9 @@ struct assignment_request {
 	/* The TSC to use if 'use' is true, otherwise automatically determine the TSC value to use. Valid range is 0 to
 	 * 7, as described in 3GPP TS 45.002. */
 	struct optional_val tsc;
+
+	/* The "speech / data indicator" from 3GPP TS 48.008 § 3.2.2.11 Channel Type (speech/data/signalling). */
+	enum gsm0808_chan_indicator ch_indctr;
 };
 
 /* State of an ongoing Assignment, while the assignment_fsm is still busy. This serves as state separation to keep the
@@ -176,7 +185,7 @@ struct assignment_fsm_data {
 	 */
 	struct assignment_request req;
 
-	bool requires_voice_stream;
+	enum gsm0808_chan_indicator ch_indctr;
 	struct channel_mode_and_rate selected_ch_mode_rate;
 
 	struct osmo_fsm_inst *fi;
@@ -273,6 +282,21 @@ struct handover {
 	struct osmo_mgcpc_ep_ci *created_ci_for_msc;
 };
 
+struct gsm_subscriber_connection;
+
+struct bscp_sccp_conn_node {
+	/* entry in (struct bsc_sccp_inst)->connections */
+	struct rb_node node;
+	/* Sigtran connection ID:
+	*  if set: Range (0..SCCP_CONN_ID_MAX) (24 bit)
+	*  if unset: SCCP_CONN_ID_UNSET (-1) if unset */
+	uint32_t conn_id;
+	/* backpointer where this sccp conn belongs to: */
+	struct gsm_subscriber_connection *gscon;
+};
+
+void bscp_sccp_conn_node_init(struct bscp_sccp_conn_node *sccp_conn, struct gsm_subscriber_connection *gscon);
+
 /* active radio connection of a mobile subscriber */
 struct gsm_subscriber_connection {
 	/* global linked list of subscriber_connections */
@@ -320,14 +344,12 @@ struct gsm_subscriber_connection {
 	/* flag to prevent multiple simultaneous ciphering commands */
 	int ciphering_handled;
 
-	/* SCCP connection associatd with this subscriber_connection */
+	/* SCCP connection associated with this subscriber_connection */
 	struct {
 		/* SCCP connection related */
 		struct bsc_msc_data *msc;
-
-		/* Sigtran connection ID */
-		int conn_id;
 		enum subscr_sccp_state state;
+		struct bscp_sccp_conn_node conn;
 	} sccp;
 
 	/* for audio handling */
@@ -376,8 +398,8 @@ struct gsm_subscriber_connection {
 
 		/* Lb interface to the SMLC: BSSMAP-LE/SCCP connection associated with this subscriber */
 		struct {
-			int conn_id;
 			enum subscr_sccp_state state;
+			struct bscp_sccp_conn_node conn;
 		} lb;
 	} lcs;
 
@@ -391,6 +413,60 @@ struct gsm_subscriber_connection {
 	} fast_return;
 
 	enum gsm0808_cause clear_cause;
+
+	/* VGCS/VBS "call controling" connection */
+	struct {
+		/* Features supported by MSC/BSC */
+		bool ff_present;
+		struct gsm0808_vgcs_feature_flags ff;
+		/* Group Call Reference IE */
+		struct gsm0808_group_callref gc_ie;
+		enum gsm0808_service_flag sf;
+		uint32_t call_ref;
+		/* Call (BSC) FSM */
+		struct osmo_fsm_inst *fi;
+		/* Current talker */
+		struct gsm_subscriber_connection *talker;
+		/* L3 info of link establihment (Talker established) */
+		struct llist_head l3_queue;
+		/* Flag and cause (Talker released) */
+		bool talker_rel;
+		uint8_t talker_cause;
+		/* Flag that states acknowledgement of the talker by MSC */
+		bool msc_ack;
+		/* List of VGCS/VBS "resource controling" connections */
+		struct llist_head chan_list;
+	} vgcs_call;
+
+	/* VGCS/VBS "resource controling" connection */
+	struct {
+		/* List entry of chan_list of "call controling" connection */
+		struct llist_head list;
+		/* Group Call Reference IE */
+		struct gsm0808_group_callref gc_ie;
+		enum gsm0808_service_flag sf;
+		uint32_t call_ref;
+		/* Channel type IE */
+		struct gsm0808_channel_type ct;
+		/* Channel mode and rate */
+		struct channel_mode_and_rate ch_mode_rate;
+		/* Cell Identifier IE */
+		struct gsm0808_cell_id ci;
+		char ci_str[16];
+		/* Assignment Requirements IE */
+		enum gsm0808_assignment_requirement ar;
+		/* Call Identifier IE */
+		uint32_t call_id;
+		/* Pointer to VGCS/VBS "call controling" gsconn */
+		struct gsm_subscriber_connection *call;
+		/* Cell (BTS) FSM */
+		struct osmo_fsm_inst *fi;
+		/* lchan to be assigned */
+		struct gsm_lchan *new_lchan;
+		/* MGW peer */
+		char msc_rtp_addr[INET6_ADDRSTRLEN];
+		uint16_t msc_rtp_port;
+	} vgcs_chan;
 };
 
 
@@ -450,9 +526,9 @@ struct gsm_abis_mo {
 	struct abis_om_obj_inst obj_inst;
 	const char *name;
 	struct gsm_nm_state nm_state;
-	struct tlv_parsed *nm_attr;
 	struct gsm_bts *bts;
 	struct osmo_fsm_inst *fi;
+	bool sw_act_rep_received;
 	bool opstart_sent;
 	bool adm_unlock_sent;
 	bool get_attr_sent;
@@ -462,6 +538,12 @@ struct gsm_abis_mo {
 	bool rsl_connect_sent;
 	bool rsl_connect_ack_received;
 	bool force_rf_lock;
+	/* vendor specific fields below */
+	union {
+		struct {
+			uint8_t obj_version;
+		} ipaccess;
+	};
 };
 
 /* Ericsson OM2000 Managed Object */
@@ -495,9 +577,9 @@ struct om2k_mo {
  *
  * These macros convert from n to the other representations:
  */
-#define ALG_A5_NR_TO_RSL(A5_N) ((A5_N) >= 0 ? (A5_N)+1 : 0)
+#define ALG_A5_NR_TO_RSL(A5_N) ((int)(A5_N) >= 0 ? (A5_N)+1 : 0)
 #define ALG_A5_NR_TO_BSSAP(A5_N) ALG_A5_NR_TO_RSL(A5_N)
-#define ALG_A5_NR_TO_PERM_ALG_BITS(A5_N) ((A5_N) >= 0 ? 1<<(A5_N) : 0)
+#define ALG_A5_NR_TO_PERM_ALG_BITS(A5_N) ((int)(A5_N) >= 0 ? 1<<(A5_N) : 0)
 
 /* Up to 16 SI2quater are multiplexed; each fits 3 EARFCNS, so the practical maximum is 3*16.
  * The real maximum that fits in a total of 16 SI2quater rest octets also depends on the bits left by other SI2quater
@@ -592,7 +674,6 @@ struct gsm_bts_trx_ts {
 	bool is_rsl_ready;
 
 	struct gsm_abis_mo mo;
-	struct tlv_parsed nm_attr;
 	uint8_t nm_chan_comb;
 	int tsc;		/* -1 == use BTS TSC */
 
@@ -806,10 +887,7 @@ struct gsm_bts *gsm_bts_by_cell_id(const struct gsm_network *net,
 extern const struct value_string gsm_chreq_descs[];
 extern const struct value_string gsm_pchant_names[];
 extern const struct value_string gsm_pchant_descs[];
-extern const struct value_string gsm_pchan_ids[];
 const char *gsm_pchan_name(enum gsm_phys_chan_config c);
-static inline const char *gsm_pchan_id(enum gsm_phys_chan_config c)
-{ return get_value_string(gsm_pchan_ids, c); }
 enum gsm_phys_chan_config gsm_pchan_parse(const char *name);
 const char *gsm_chreq_name(enum gsm_chreq_reason_t c);
 char *gsm_ts_name(const struct gsm_bts_trx_ts *ts);
@@ -819,6 +897,8 @@ void gsm_abis_mo_reset(struct gsm_abis_mo *mo);
 void gsm_mo_init(struct gsm_abis_mo *mo, struct gsm_bts *bts,
 		 uint8_t obj_class, uint8_t p1, uint8_t p2, uint8_t p3);
 
+struct gsm_abis_mo *gsm_objclass2mo(struct gsm_bts *bts, uint8_t obj_class,
+				    const struct abis_om_obj_inst *obj_inst);
 struct gsm_nm_state *
 gsm_objclass2nmstate(struct gsm_bts *bts, uint8_t obj_class,
 		 const struct abis_om_obj_inst *obj_inst);
@@ -863,6 +943,19 @@ struct all_allocated {
 	struct osmo_time_cc static_tch;
 };
 
+struct bsc_sccp_inst {
+	struct osmo_sccp_instance *sccp; /* backpointer */
+	/* rbtree root of 'struct bscp_sccp_conn_node' in this instance, ordered by conn_id */
+	struct rb_root connections;
+	uint32_t next_id; /* next id to allocate */
+};
+
+struct bsc_sccp_inst *bsc_sccp_inst_alloc(void *ctx);
+uint32_t bsc_sccp_inst_next_conn_id(struct bsc_sccp_inst *bsc_sccp);
+int bsc_sccp_inst_register_gscon(struct bsc_sccp_inst *bsc_sccp, struct bscp_sccp_conn_node *sccp_conn);
+void bsc_sccp_inst_unregister_gscon(struct bsc_sccp_inst *bsc_sccp, struct bscp_sccp_conn_node *sccp_conn);
+struct gsm_subscriber_connection *bsc_sccp_inst_get_gscon_by_conn_id(const struct bsc_sccp_inst *bsc_sccp, uint32_t conn_id);
+
 struct gsm_network {
 	struct osmo_plmn_id plmn;
 
@@ -899,10 +992,10 @@ struct gsm_network {
 
 	/* msc configuration */
 	struct llist_head mscs;
-	uint8_t mscs_round_robin_next_nr;
+	unsigned int mscs_round_robin_next_nr;
 	/* Emergency calls potentially select a different set of MSCs, so to not mess up the normal round-robin
 	 * behavior, emergency calls need a separate round-robin counter. */
-	uint8_t mscs_round_robin_next_emerg_nr;
+	unsigned int mscs_round_robin_next_emerg_nr;
 
 	/* rf ctl related bits */
 	int mid_call_timeout;
@@ -928,13 +1021,8 @@ struct gsm_network {
 	 * OsmoMSC, this should be tied to the location area code (LAC). */
 	struct gsm_tz tz;
 
-	/* List of all struct bsc_subscr used in libbsc. This llist_head is
-	 * allocated so that the llist_head pointer itself can serve as a
-	 * talloc context (useful to not have to pass the entire gsm_network
-	 * struct to the bsc_subscr_* API, and for bsc_susbscr unit tests to
-	 * not require gsm_data.h). In an MSC-without-BSC environment, this
-	 * pointer is NULL to indicate absence of a bsc_subscribers list. */
-	struct llist_head *bsc_subscribers;
+	/* Keeps track of struct bsc_subcr. */
+	struct bsc_subscr_store *bsc_subscribers;
 
 	/* Timer for periodic channel load measurements to maintain each BTS's T3122. */
 	struct osmo_timer_list t3122_chan_load_timer;
@@ -977,6 +1065,11 @@ struct gsm_network {
 
 	struct chan_counts chan_counts;
 	struct all_allocated all_allocated;
+
+	/* PCU socket state */
+	char *pcu_sock_path;
+	unsigned int pcu_sock_wqueue_len_max;
+	struct pcu_sock_state *pcu_state;
 };
 
 struct gsm_audio_support {
@@ -984,6 +1077,8 @@ struct gsm_audio_support {
                 ver : 7;
 };
 
+int gsm_audio_support_cmp(const struct gsm_audio_support *a, const struct gsm_audio_support *b);
+
 extern void talloc_ctx_init(void *ctx_root);
 
 enum gsm_bts_type parse_btstype(const char *arg);
@@ -1020,8 +1115,6 @@ enum gsm_phys_chan_config gsm_pchan_by_lchan_type(enum gsm_chan_t type);
 enum gsm48_rr_cause bsc_gsm48_rr_cause_from_gsm0808_cause(enum gsm0808_cause c);
 enum gsm48_rr_cause bsc_gsm48_rr_cause_from_rsl_cause(uint8_t c);
 
-int bsc_sccp_inst_next_conn_id(struct osmo_sccp_instance *sccp);
-
 /* Interference Measurement Parameters */
 struct gsm_interf_meas_params {
 	/* Intave: Interference Averaging period (see 3GPP TS 45.008, table A.1) */
